// Copyright 2022 schukai GmbH
// SPDX-License-Identifier: AGPL-3.0

package configuration

import (
	"bytes"
	"encoding/json"
	"github.com/imdario/mergo"
	"github.com/magiconair/properties"
	"github.com/pelletier/go-toml/v2"
	"gopkg.in/yaml.v3"
	"io"
	"os"
	"path"
)

func importJson[C any](config *C, reader io.Reader) error {
	decoder := json.NewDecoder(reader)
	return decoder.Decode(config)
}

func importYaml[C any](config *C, reader io.Reader) error {
	decoder := yaml.NewDecoder(reader)
	return decoder.Decode(config)
}

func importToml[C any](config *C, reader io.Reader) error {
	decoder := toml.NewDecoder(reader)
	return decoder.Decode(config)
}

func importProperties[C any](config *C, reader io.Reader) error {

	buf := &bytes.Buffer{}
	buf.ReadFrom(reader)
	b := buf.Bytes()

	p, err := properties.Load(b, properties.UTF8)
	if err != nil {
		return err
	}

	return p.Decode(config)

}

func (s *Settings[C]) importStream(r reader) {
	var c C
	var err error

	err = nil // reset error

	x := r.reader
	switch r.format {
	case Json:
		err = importJson(&c, x)
	case Yaml:
		err = importYaml(&c, x)
	case Toml:
		err = importToml(&c, x)
	case Properties:
		err = importProperties(&c, x)
	default:
		err = newFormatNotSupportedError(r.format)
	}

	if err == nil {
		s.importCounter++
	} else {
		s.errors = append(s.errors, err)
	}

	if err := mergo.Merge(&s.config, c); err != nil {
		s.errors = append(s.errors, err)
	}
}

func (s *Settings[C]) importStreams() {
	for _, r := range s.stream.readers {
		s.importStream(r)
	}
}

func (s *Settings[C]) importFiles() {

	s.fileWatch.Lock()
	defer s.fileWatch.Unlock()

	// new files may have been added
	s.fileWatch.watchList = make(map[string]string)

	for _, d := range s.files.directories {
		fn := path.Join(d, s.files.name+s.files.format.Extension())

		f, err := s.files.fs.Open(fn)

		if os.IsNotExist(err) {
			f.Close()
			continue
		}

		r := (io.Reader)(f)
		s.importStream(reader{s.files.format, r})
		f.Close()

		s.fileWatch.watchList[fn] = fn

	}

	for _, f := range s.files.files {
		r, err := s.files.fs.Open(f.path)

		if err != nil {
			s.errors = append(s.errors, err)
			r.Close()
			continue
		}
		s.importStream(reader{f.format, r})
		r.Close()
		s.fileWatch.watchList[f.path] = f.path
	}

}

func (s *Settings[C]) Import() *Settings[C] {

	s.Lock()
	defer s.Unlock()

	defaults := s.config

	var n C
	s.config = n

	s.importCounter = 0

	s.importFiles()
	s.importStreams()

	if err := mergo.Merge(&s.config, defaults); err != nil {
		s.errors = append(s.errors, err)
	}

	if s.importCounter == 0 {
		s.errors = append(s.errors, NoFileImportedError)
	}

	x := s.config
	s.config = defaults
	s.setConfigInternal(x, false)

	return s
}
